HT_TP_GLOBAL_SCOPED_INT(ok_1);

HT_TP_GLOBAL_SCOPED_INT(1_not_ok);

HT_TP_GLOBAL_SCOPED_INT(not_ok)

HT_TP_GLOBAL_SCOPED_INT(_ok_2);

HT_TP_GLOBAL_SCOPED_STRING(not_ok);

HT_TP_GLOBAL_SCOPED_INT(_ok_2);

HT_TP_GLOBAL_SCOPED_INT("Not ok");
